"""
This module defines the |OME| for the non-trivial matching conditions in the
|VFNS| evolution.
"""

import numba as nb
import numpy as np

from . import as1, as2, as3


@nb.njit(cache=True)
def A_singlet(matching_order, n, sx, nf, L, is_msbar, sx_ns=None):
    r"""
    Computes the tower of the singlet |OME|.

    Parameters
    ----------
        matching_order : tuple(int,int)
            perturbative matching_order
        n : complex
            Mellin variable
        sx : list
            singlet like harmonic sums cache
        nf: int
            number of active flavor below threshold
        L : float
            :math:``\ln(\mu_F^2 / m_h^2)``
        is_msbar: bool
            add the |MSbar| contribution
        sx_ns : list
            non-singlet like harmonic sums cache

    Returns
    -------
        A_singlet : numpy.ndarray
            singlet |OME|

    See Also
    --------
        ekore.matching_conditions.nlo.A_singlet_1 : :math:`A^{S,(1)}(N)`
        ekore.matching_conditions.nlo.A_hh_1 : :math:`A_{HH}^{(1)}(N)`
        ekore.matching_conditions.nlo.A_gh_1 : :math:`A_{gH}^{(1)}(N)`
        ekore.matching_conditions.nnlo.A_singlet_2 : :math:`A_{S,(2)}(N)`
    """
    A_s = np.zeros((matching_order[0], 3, 3), np.complex_)
    if matching_order[0] >= 1:
        A_s[0] = as1.A_singlet(n, sx, L)
    if matching_order[0] >= 2:
        A_s[1] = as2.A_singlet(n, sx, L, is_msbar)
    if matching_order[0] >= 3:
        A_s[2] = as3.A_singlet(n, sx, sx_ns, nf, L)
    return A_s


@nb.njit(cache=True)
def A_non_singlet(matching_order, n, sx, nf, L):
    r"""
    Computes the tower of the non-singlet |OME|

    Parameters
    ----------
        matching_order : tuple(int,int)
            perturbative matching_order
        n : complex
            Mellin variable
        sx : list
            harmonic sums cache
        nf: int
            number of active flavor below threshold
        L : float
            :math:``\ln(\mu_F^2 / m_h^2)``

    Returns
    -------
        A_non_singlet : numpy.ndarray
            non-singlet |OME|

    See Also
    --------
        ekore.matching_conditions.nlo.A_hh_1 : :math:`A_{HH}^{(1)}(N)`
        ekore.matching_conditions.nnlo.A_ns_2 : :math:`A_{qq,H}^{NS,(2)}`
    """
    A_ns = np.zeros((matching_order[0], 2, 2), np.complex_)
    if matching_order[0] >= 1:
        A_ns[0] = as1.A_ns(n, sx, L)
    if matching_order[0] >= 2:
        A_ns[1] = as2.A_ns(n, sx, L)
    if matching_order[0] >= 3:
        A_ns[2] = as3.A_ns(n, sx, nf, L)
    return A_ns



r"""The polarized, space-like |OME|."""

import numba as nb


@nb.njit(cache=True)
def A_non_singlet(_matching_order, _n, _sx, _nf, _L):
    raise NotImplementedError("Polarised, space-like is not yet implemented")


@nb.njit(cache=True)
def A_singlet(_matching_order, _n, _sx, _nf, _L, _is_msbar, _sx_ns=None):
    raise NotImplementedError("Polarised, space-like is not yet implemented")



r"""The unpolarized, time-like Altarelli-Parisi splitting kernels.

Normalization is given by

.. math::
    \mathbf{P}(x) = \sum\limits_{j=0} a_s^{j+1} \mathbf P^{(j)}(x)

with :math:`a_s = \frac{\alpha_S(\mu^2)}{4\pi}`.
"""

import numba as nb


@nb.njit(cache=True)
def gamma_ns(_order, _mode, _n, _nf):
    raise NotImplementedError("Polarised is not yet implemented")


@nb.njit(cache=True)
def gamma_singlet(_order, _n, _nf):
    raise NotImplementedError("Polarised is not yet implemented")



"""Structures to hold runcards information.

All energy scales in the runcards should be saved linearly, not the squared
value, for consistency.
Squares are consistenly taken inside.

"""
from dataclasses import dataclass
from math import nan
from typing import Optional, Union

import numpy as np
import numpy.typing as npt

from .. import basis_rotation as br
from .. import interpolation
from .. import version as vmod
from .dictlike import DictLike
from .types import (
    CouplingsRef,
    EvolutionMethod,
    FlavorsNumber,
    HeavyQuarkMasses,
    IntrinsicFlavors,
    InversionMethod,
    MatchingScales,
    Order,
    QuarkMassSchemes,
    RawCard,
    ScaleVariationsMethod,
    T,
)


@dataclass
class TheoryCard(DictLike):
    """Represent theory card content."""

    order: Order
    """Perturbative order tuple, ``(QCD, QED)``."""
    couplings: CouplingsRef
    """Couplings configuration."""
    num_flavs_init: Optional[FlavorsNumber]
    r"""Number of active flavors at fitting scale.

    I.e. :math:`n_{f,\text{ref}}(\mu^2_0)`, formerly called ``nf0``.

    """
    num_flavs_max_pdf: FlavorsNumber
    """Maximum number of quark PDFs."""
    intrinsic_flavors: IntrinsicFlavors
    """List of intrinsic quark PDFs."""
    quark_masses: HeavyQuarkMasses
    """List of heavy quark masses."""
    quark_masses_scheme: QuarkMassSchemes
    """Scheme used to specify heavy quark masses."""
    matching: MatchingScales
    """Matching scale of heavy quark masses"""
    xif: float
    """Ratio between factorization scale and process scale."""


@dataclass
class Debug(DictLike):
    """Debug configurations."""

    skip_singlet: bool = False
    """Whether to skip QCD singlet computation."""
    skip_non_singlet: bool = False
    """Whether to skip QCD non-singlet computation."""


@dataclass
class Configs(DictLike):
    """Solution specific configurations."""

    evolution_method: EvolutionMethod
    """Evolution mode."""
    ev_op_max_order: Order
    """Maximum order to use in U matrices expansion.
    Used only in ``perturbative`` solutions.
    """
    ev_op_iterations: int
    """Number of intervals in which to break the global path."""
    interpolation_polynomial_degree: int
    """Degree of elements of the intepolation polynomial basis."""
    interpolation_is_log: bool
    r"""Whether to use polynomials in :math:`\log(x)`.
    If `false`, polynomials are in :math:`x`.
    """
    polarized: bool
    """If `true` do polarized evolution."""
    time_like: bool
    """If `true` do time-like evolution."""
    scvar_method: Optional[ScaleVariationsMethod]
    """"""
    inversion_method: Optional[InversionMethod]
    """Which method to use for backward matching conditions."""
    n_integration_cores: int = 1
    """Number of cores used to parallelize integration."""


@dataclass
class Rotations(DictLike):
    """Rotations related configurations.

    Here "Rotation" is intended in a broad sense: it includes both rotations in
    flavor space (labeled with suffix `pids`) and in :math:`x`-space (labeled
    with suffix `grid`).
    Rotations in :math:`x`-space correspond to reinterpolate the result on a
    different basis of polynomials.

    """

    xgrid: interpolation.XGrid
    """Momentum fraction internal grid."""
    pids: npt.NDArray
    """Array of integers, corresponding to internal PIDs."""
    _targetgrid: Optional[interpolation.XGrid] = None
    _inputgrid: Optional[interpolation.XGrid] = None
    _targetpids: Optional[npt.NDArray] = None
    _inputpids: Optional[npt.NDArray] = None

    def __post_init__(self):
        """Adjust types when loaded from serialized object."""
        for attr in ("xgrid", "_inputgrid", "_targetgrid"):
            value = getattr(self, attr)
            if value is None:
                continue
            if isinstance(value, (np.ndarray, list)):
                setattr(self, attr, interpolation.XGrid(value))
            elif not isinstance(value, interpolation.XGrid):
                setattr(self, attr, interpolation.XGrid.load(value))

    @property
    def inputpids(self) -> npt.NDArray:
        """Provide pids expected on the input PDF."""
        if self._inputpids is None:
            return self.pids
        return self._inputpids

    @inputpids.setter
    def inputpids(self, value):
        self._inputpids = value

    @property
    def targetpids(self) -> npt.NDArray:
        """Provide pids corresponding to the output PDF."""
        if self._targetpids is None:
            return self.pids
        return self._targetpids

    @targetpids.setter
    def targetpids(self, value):
        self._targetpids = value

    @property
    def inputgrid(self) -> interpolation.XGrid:
        """Provide :math:`x`-grid expected on the input PDF."""
        if self._inputgrid is None:
            return self.xgrid
        return self._inputgrid

    @inputgrid.setter
    def inputgrid(self, value: interpolation.XGrid):
        self._inputgrid = value

    @property
    def targetgrid(self) -> interpolation.XGrid:
        """Provide :math:`x`-grid corresponding to the output PDF."""
        if self._targetgrid is None:
            return self.xgrid
        return self._targetgrid

    @targetgrid.setter
    def targetgrid(self, value: interpolation.XGrid):
        self._targetgrid = value


@dataclass
class OperatorCard(DictLike):
    """Operator Card info."""

    mu0: float
    """Initial scale."""

    # collections
    configs: Configs
    """Specific configuration to be used during the calculation of these operators."""
    debug: Debug
    """Debug configurations."""
    rotations: Rotations
    """Rotations configurations.

    The operator card will only contain the interpolation xgrid and the pids.

    """

    # TODO: drop legacy compatibility, only linear scales in runcards, such
    # that we will always avoid taking square roots, and we are consistent with
    # the other scales
    _mugrid: Optional[npt.NDArray] = None
    _mu2grid: Optional[npt.NDArray] = None
    """Array of final scales."""

    # optional
    eko_version: str = vmod.__version__
    """Version of EKO package first used for generation."""

    # a few properties, for ease of use and compatibility
    @property
    def mu20(self):
        """Squared value of initial scale."""
        return self.mu0**2

    @property
    def mugrid(self):
        """Only setter enabled, access only to :attr:`mu2grid`."""
        raise ValueError("Use mu2grid")

    @mugrid.setter
    def mugrid(self, value):
        """Set scale grid with linear values."""
        self._mugrid = value
        self._mu2grid = None

    @property
    def mu2grid(self):
        """Grid of squared final scales."""
        if self._mugrid is not None:
            return self._mugrid**2
        if self._mu2grid is not None:
            return self._mu2grid

        raise RuntimeError("Mu2 grid has not been initialized")

    @mu2grid.setter
    def mu2grid(self, value):
        """Set scale grid with quadratic values."""
        self._mugrid = None
        self._mu2grid = value


Card = Union[TheoryCard, OperatorCard]


@dataclass
class Legacy:
    """Upgrade legacy runcards."""

    theory: RawCard
    operator: RawCard

    MOD_EV2METHOD = {
        "EXA": "iterate-exact",
        "EXP": "iterate-expanded",
        "TRN": "truncated",
    }
    HEAVY = "cbt"

    @staticmethod
    def fallback(*args: T, default: Optional[T] = None) -> T:
        """Return the first not None argument."""
        for arg in args:
            if arg is not None:
                return arg

        if default is None:
            return args[-1]
        return default

    @property
    def new_theory(self):
        """Build new format theory runcard."""
        old = self.theory
        new = {}

        def heavies(pattern: str):
            return {q: old[pattern % q] for q in self.HEAVY}

        new["order"] = [old["PTO"] + 1, old["QED"]]
        alphaem = self.fallback(old.get("alphaqed"), old.get("alphaem"), default=0.0)
        if "QrefQED" not in old:
            qedref = nan
        else:
            qedref = old["QrefQED"]
        new["couplings"] = dict(
            alphas=(old["alphas"], old["Qref"]),
            alphaem=(alphaem, qedref),
            num_flavs_ref=old["nfref"],
            max_num_flavs=old["MaxNfAs"],
        )
        new["num_flavs_init"] = old["nf0"]
        new["num_flavs_max_pdf"] = old["MaxNfPdf"]
        intrinsic = []
        for idx, q in enumerate(self.HEAVY):
            if old.get(f"i{q}".upper()) == 1:
                intrinsic.append(idx + 4)
        new["intrinsic_flavors"] = intrinsic
        new["matching"] = heavies("k%sThr")
        new["quark_masses_scheme"] = old["HQ"]
        ms = heavies("m%s")
        mus = heavies("Qm%s")
        if old["HQ"] == "POLE":
            new["quark_masses"] = {q: (ms[q], nan) for q in self.HEAVY}
        elif old["HQ"] == "MSBAR":
            new["quark_masses"] = {q: (ms[q], mus[q]) for q in self.HEAVY}
        else:
            raise ValueError()

        new["xif"] = old["XIF"]

        return TheoryCard.from_dict(new)

    @property
    def new_operator(self):
        """Build new format operator runcard."""
        old = self.operator
        old_th = self.theory
        new = {}

        new["mu0"] = old_th["Q0"]
        new["_mu2grid"] = old["Q2grid"]

        new["configs"] = {}
        evmod = old_th["ModEv"]
        new["configs"]["evolution_method"] = self.MOD_EV2METHOD.get(evmod, evmod)
        new["configs"]["inversion_method"] = old_th.get(
            "backward_inversion", "expanded"
        )
        new["configs"]["scvar_method"] = old_th.get("ModSV", "expanded")
        for k in (
            "interpolation_polynomial_degree",
            "interpolation_is_log",
            "ev_op_iterations",
            "n_integration_cores",
            "polarized",
            "time_like",
        ):
            new["configs"][k] = old[k]
        max_order = old["ev_op_max_order"]
        if isinstance(max_order, int):
            new["configs"]["ev_op_max_order"] = [max_order, old_th["QED"]]

        new["debug"] = {}
        lpref = len("debug_")
        for k in ("debug_skip_non_singlet", "debug_skip_singlet"):
            new["debug"][k[lpref:]] = old[k]

        new["rotations"] = {}
        new["rotations"]["pids"] = old.get("pids", br.flavor_basis_pids)
        new["rotations"]["xgrid"] = old["interpolation_xgrid"]
        for basis in ("inputgrid", "targetgrid", "inputpids", "targetpids"):
            new["rotations"][f"_{basis}"] = old[basis]

        return OperatorCard.from_dict(new)


def update(theory: Union[RawCard, TheoryCard], operator: Union[RawCard, OperatorCard]):
    """Update legacy runcards.

    This function is mainly defined for compatibility with the old interface.
    Prefer direct usage of :class:`Legacy` in new code.

    Consecutive applications of this function yield identical results::

        cards = update(theory, operator)
        assert update(*cards) == cards

    """
    if isinstance(theory, TheoryCard) or isinstance(operator, OperatorCard):
        # if one is not a dict, both have to be new cards
        assert isinstance(theory, TheoryCard)
        assert isinstance(operator, OperatorCard)
        return theory, operator

    cards = Legacy(theory, operator)
    return cards.new_theory, cards.new_operator



"""Common type definitions, only used for static analysis.

Unfortunately, Python has still some discrepancies between runtime classes and
type hints, so it is better not to mix dataclasses and generics.

E.g. before it was implemented::

    @dataclasses.dataclass
    class RunningReference(DictLike, Generic[Quantity]):
        value: Quantity
        scale: float

but in this way it is not possible to determine that ``RunningReference`` is
subclassing ``DictLike``, indeed::

    inspec.isclass(RunningReference)  # False
    issubclass(RunningReference, DictLike)  # raise an error, since
                                            # RunningReference is not a class

Essentially classes can be used for type hints, but types are not all classes,
especially when they involve generics.

For this reason I prefer the less elegant dynamic generation, that seems to
preserve type hints.

"""
import dataclasses
import enum
import typing
from math import isnan
from typing import Any, Dict, Optional

from .dictlike import DictLike

T = typing.TypeVar("T")


def reference_running(quantity: typing.Type[typing.Union[int, float]]):
    """Generate running quantities reference point classes.

    The motivation for dynamic generation is provided in module docstring.

    """

    @dataclasses.dataclass
    class ReferenceRunning(DictLike):
        value: quantity
        scale: float

    return ReferenceRunning


IntRef = reference_running(int)
FloatRef = reference_running(float)

Order = typing.Tuple[int, int]
FlavorsNumber = int
FlavorIndex = int
IntrinsicFlavors = typing.List[FlavorIndex]


@dataclasses.dataclass
class CouplingsRef(DictLike):
    """Reference values for coupling constants."""

    alphas: FloatRef
    alphaem: FloatRef
    max_num_flavs: FlavorsNumber
    num_flavs_ref: Optional[FlavorsNumber]
    r"""Number of active flavors at strong coupling reference scale.

    I.e. :math:`n_{f,\text{ref}}(\mu^2_{\text{ref}})`, formerly called
    ``nfref``.

    """

    def __post_init__(self):
        """Validate couplings.

        If they are both running, they have to be defined at the same scale.

        Usually :attr:`alphaem` is not running, thus its scale is set to nan.

        """
        assert self.alphas.scale == self.alphaem.scale or isnan(self.alphaem.scale)

    @property
    def values(self):
        """Collect only couplings values."""
        return (self.alphas.value, self.alphaem.value)


def heavy_quark(quarkattr):
    """Generate heavy quark properties container classes.

    The motivation for dynamic generation is provided in module docstring.

    """

    @dataclasses.dataclass
    class HeavyQuarks(DictLike):
        """Access heavy quarks attributes by name."""

        c: quarkattr
        """Charm quark."""
        b: quarkattr
        """Bottom quark."""
        t: quarkattr
        """Top quark."""

        def __getitem__(self, key: int):
            """Allow access by index.

            Consequently it allows iteration and containing check.

            """
            return getattr(self, "cbt"[key])

    return HeavyQuarks


QuarkMassRef = FloatRef
HeavyQuarkMasses = heavy_quark(QuarkMassRef)
MatchingScale = float
MatchingScales = heavy_quark(MatchingScale)


# TODO: upgrade all the following to StrEnum, new in py3.11
# with that, it is possible to replace all non-alias right sides with calls to
# enum.auto()


class QuarkMassSchemes(enum.Enum):
    """Scheme to define heavy quark masses."""

    MSBAR = "msbar"
    POLE = "pole"


class EvolutionMethod(enum.Enum):
    """DGLAP solution method."""

    ITERATE_EXACT = "iterate-exact"
    ITERATE_EXPANDED = "iterate-expanded"
    PERTURBATIVE_EXACT = "perturbative-exact"
    PERTURBATIVE_EXPANDED = "perturbative-expanded"
    TRUNCATED = "truncated"
    ORDERED_TRUNCATED = "ordered-truncated"
    DECOMPOSE_EXACT = "decompose-exact"
    DECOMPOSE_EXPANDED = "decompose-expanded"


class CouplingEvolutionMethod(enum.Enum):
    """Beta functions solution method."""

    EXACT = "exact"
    EXPANDED = "expanded"


class ScaleVariationsMethod(enum.Enum):
    """Method used to account for factorization scale variation."""

    EXPONENTIATED = "exponentiated"
    EXPANDED = "expanded"


class InversionMethod(enum.Enum):
    """Method used to invert the perturbative matching conditions."""

    EXACT = "exact"
    EXPANDED = "expanded"


RawCard = Dict[str, Any]
